/*
 * This file is part of Sponge, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.spongepowered.common.item.inventory.lens.impl.collections;

import static com.google.common.base.Preconditions.*;

import net.minecraft.inventory.IInventory;
import net.minecraft.item.ItemStack;
import org.spongepowered.api.item.inventory.Inventory;
import org.spongepowered.api.item.inventory.Slot;
import org.spongepowered.common.item.inventory.adapter.InventoryAdapter;
import org.spongepowered.common.item.inventory.adapter.impl.SlotCollectionIterator;
import org.spongepowered.common.item.inventory.adapter.impl.slots.SlotAdapter;
import org.spongepowered.common.item.inventory.lens.Fabric;
import org.spongepowered.common.item.inventory.lens.Lens;
import org.spongepowered.common.item.inventory.lens.SlotProvider;
import org.spongepowered.common.item.inventory.lens.impl.slots.SlotLensImpl;
import org.spongepowered.common.item.inventory.lens.slots.SlotLens;

import java.util.ArrayList;
import java.util.List;

public class SlotCollection extends DynamicLensCollectionImpl<IInventory, ItemStack> implements SlotProvider<IInventory, ItemStack> {
    
    public static class Builder {
        
        private List<Class<? extends SlotAdapter>> slotTypes = new ArrayList<Class<? extends SlotAdapter>>();
        
        public Builder add() {
            return this.add(SlotAdapter.class);
        }
        
        public Builder add(Class<? extends SlotAdapter> type) {
            this.slotTypes.add(checkNotNull(type));
            return this;
        }
        
        public Builder add(int count) {
            return this.add(count, SlotAdapter.class);
        }
        
        public Builder add(int count, Class<? extends SlotAdapter> type) {
            checkNotNull(type);
            for (int i = 0; i < count; i++) {
                this.slotTypes.add(type);
            }
            
            return this;
        }
        
        public int size() {
            return this.slotTypes.size();
        }
        
        Class<? extends Inventory> getType(int index) {
            return this.slotTypes.get(index);
        }
        
        public SlotCollection build() {
            return new SlotCollection(this.size(), this);
        }

    }
    
    private Builder builder;

    public SlotCollection(int size) {
        this(size, null);
    }

    SlotCollection(int size, Builder builder) {
        super(size);
        this.builder = builder;
        this.populate();
    }
    
    private void populate() {
        for (int index = 0; index < this.size(); index++) {
            this.lenses[index] = this.createSlotLens(index);
        }
    }

    protected SlotLens<IInventory, ItemStack> createSlotLens(int slotIndex) {
        return new SlotLensImpl(slotIndex, this.getSlotAdapterType(slotIndex));
    }

    protected Class<? extends Inventory> getSlotAdapterType(int slotIndex) {
        try {
            return this.builder != null ? this.builder.getType(slotIndex) : SlotAdapter.class;
        } catch (IndexOutOfBoundsException ex) {
            return SlotAdapter.class;
        }
    }

    @Override
    public SlotLens<IInventory, ItemStack> getSlot(int index) {
        return (SlotLens<IInventory, ItemStack>) this.get(index);
    }

    public Iterable<Slot> getIterator(InventoryAdapter<IInventory, ItemStack> adapter) {
        return this.getIterator(adapter, adapter.getInventory(), adapter.getRootLens());
    }
    
    public Iterable<Slot> getIterator(Inventory parent, InventoryAdapter<IInventory, ItemStack> adapter) {
        return this.getIterator(parent, adapter.getInventory(), adapter.getRootLens());
    }
    
    public Iterable<Slot> getIterator(Inventory parent, Fabric<IInventory> inv, Lens<IInventory, ItemStack> lens) {
        return new SlotCollectionIterator(parent, inv, lens, this);
    }

}
